/*
 * (c) Copyright 2018-2020 DESY
 *
 * This file is part of FS-DS detector software.
 *
 * This software is free: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 *************************************************************************
 *     Author: Yuelong Yu <yuelong.yu@desy.de>
 */


#include "AGIPDComm.h"

namespace AGIPDCtrlNS
{
    AGIPDComm::AGIPDComm()
        :m_fileop(new FileReader()),
         m_cmd(COMMAND_HEADER_LENGTH,0x00),
         m_response(CHAR_BUFFER_SIZE,0x00),
         m_tag(0)
    {
        LOG_TRACE(__FUNCTION__);
    }
    
    AGIPDComm::~AGIPDComm()
    {
        LOG_TRACE(__FUNCTION__);
    }
    
    bool AGIPDComm::Init(sptr_network& tcp,string log_path)
    {
        LOG_TRACE(__FUNCTION__);
        
        m_tcp = tcp;
        m_logpath = log_path;
        
        return true;
    }

    string AGIPDComm::GetFirmwareVersion()
    {
        LOG_TRACE(__FUNCTION__);

        return string("unknown");
    }

   
    RESPONSE AGIPDComm::GetApplicationLog(string& log)
    {
        LOG_TRACE(__FUNCTION__);

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x80;
        cmd[1] = 0x03; // command code
        cmd[2] = 0x00; // to ZYNC
        
        // send mac address
        cmd[3] = GetSequenceTag();
        cmd[5] = 0x00;

        PrintCmd(cmd);
        
        m_tcp->SendData(cmd);

        vector<char> response(6,0x00);
        RESPONSE rc = CheckResponseLog(response);
        
        string content(response.begin(),response.end());
        log = content;

        //if(rc == SC)
        WriteLogToFile("application",content);
        
        return rc;
    }

    RESPONSE AGIPDComm::GetCompilationLog(string& log)
    {
         LOG_TRACE(__FUNCTION__);

         vector<uint8> cmd = m_cmd;
         cmd[0] = 0x80;
         cmd[1] = 0x02; // command code
         cmd[2] = 0x00; // to ZYNC


         // get compilation log
         cmd[3] = GetSequenceTag();
         cmd[5] = 0x00;

         PrintCmd(cmd);
        
         m_tcp->SendData(cmd);

         vector<char> response(6,0x00);
         RESPONSE rc = CheckResponseLog(response);

         string content(response.begin(),response.end());
         log = content;

         //if(rc == SC)
         WriteLogToFile("compilation",content);
         
         return rc;
    }
    
    RESPONSE AGIPDComm::Set10GElinkMAC(string mac)
    {
        LOG_TRACE(__FUNCTION__);

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x81;
        cmd[1] = 0x00; // command code
        cmd[2] = 0x00; // to ZYNC


        // send mac address
        cmd[3] = GetSequenceTag();
        cmd[5] = 0x06;
        
        vector<uint32> macaddr(6,0);
        sscanf(mac.c_str(),
               "%x:%x:%x:%x:%x:%x",
               &macaddr[0],&macaddr[1],&macaddr[2],
               &macaddr[3],&macaddr[4],&macaddr[5]);

        for(auto val : macaddr)
            cmd.push_back(static_cast<uint8>(val));

        PrintCmd(cmd);
        
        m_tcp->SendData(cmd);
        
        vector<char> response(6,0x00);
        return CheckResponse(response);
    }

    RESPONSE AGIPDComm::Set10GElinkIP(string ip)
    {
        LOG_TRACE(__FUNCTION__);

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x81;
        cmd[1] = 0x01; //command code
        cmd[2] = 0x00; // to ZYNC
        
        // send ip address
        cmd[3] = GetSequenceTag();
        cmd[5] = 0x04;
        
        vector<uint32> ipaddr(4,0);
        sscanf(ip.c_str(),
               "%d.%d.%d.%d",
               &ipaddr[0],&ipaddr[1],
               &ipaddr[2],&ipaddr[3]);

        for(auto val : ipaddr)
            cmd.push_back(static_cast<uint8>(val));

        PrintCmd(cmd);
        
        m_tcp->SendData(cmd);

        vector<char> response(6,0x00);
        return CheckResponse(response);
    }
    
    
    RESPONSE AGIPDComm::SetHighVoltage(uint16 high_voltage)
    {
        LOG_TRACE(__FUNCTION__);

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x81;
        cmd[1] = 0x04; // command code
        cmd[2] = 0x00; // to ZYNC

        cmd[3] = GetSequenceTag();
        cmd[5] = 0x02;

        // high voltage
        cmd.push_back(static_cast<uint8>((high_voltage >> 8) & 0xff));
        cmd.push_back(static_cast<uint8>(high_voltage & 0xff));

        PrintCmd(cmd);
        
        m_tcp->SendData(cmd);

        vector<char> response(6,0x00);
        return CheckResponse(response);        
    }
    
    RESPONSE AGIPDComm::GetHighVoltage(uint16& high_voltage)
    {
        LOG_TRACE(__FUNCTION__);

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x80;
        cmd[1] = 0x05; // command code
        cmd[2] = 0x00; // to ZYNC

        cmd[3] = GetSequenceTag();
        cmd[5] = 0x00;
        
        PrintCmd(cmd);
        
        m_tcp->SendData(cmd);

        vector<char> response(8,0x00);
        RESPONSE rc = CheckResponse(response);

        if(rc == SC)
            high_voltage = (((int16)response[6] << 8) | ((int16)response[7]));
        else
            high_voltage = 0;

        return rc;
    }

    RESPONSE AGIPDComm::SetHighVoltageSlewRate(uint16 hv_slewrate)
    {
        LOG_TRACE(__FUNCTION__);

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x81;
        cmd[1] = 0x06; // command code
        cmd[2] = 0x00; // to ZYNC

        cmd[3] = GetSequenceTag();
        cmd[5] = 0x02;

        // high voltage
        cmd.push_back(static_cast<uint8>((hv_slewrate >> 8) & 0xff));
        cmd.push_back(static_cast<uint8>(hv_slewrate & 0xff));

        PrintCmd(cmd);
        
        m_tcp->SendData(cmd);

        vector<char> response(6,0x00);
        return CheckResponse(response);
    }

    RESPONSE AGIPDComm::GetHighVoltageSlewRate(uint16& hv_slewrate)
    {
        LOG_TRACE(__FUNCTION__);

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x80;
        cmd[1] = 0x05; // command code
        cmd[2] = 0x00; // to ZYNC

        cmd[3] = GetSequenceTag();
        cmd[5] = 0x00;
        
        PrintCmd(cmd);
        
        m_tcp->SendData(cmd);


        vector<char> response(8,0x00);
        RESPONSE rc = CheckResponse(response);

        if(rc == SC)
            hv_slewrate = (((int16)response[6] << 8) | ((int16)response[7]));
        else
            hv_slewrate = 0;

        return rc;
    }

    RESPONSE AGIPDComm::ASICPowerUP(int8 asic_mask)
    {
        LOG_TRACE(__FUNCTION__);

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x80;
        cmd[1] = 0x07; // command code
        cmd[2] = 0x00; // to ZYNC

        cmd[3] = GetSequenceTag();
        cmd[5] = 0x01;

        cmd.push_back(static_cast<uint8>(asic_mask));
        PrintCmd(cmd);
        
        m_tcp->SendData(cmd);

        vector<char> response(6,0x00);
        return CheckResponse(response);
    }

    RESPONSE AGIPDComm::ASICPowerDown(int8 asic_mask)
    {
        LOG_TRACE(__FUNCTION__);

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x80;
        cmd[1] = 0x08; // command code
        cmd[2] = 0x00; // to ZYNC

        cmd[3] = GetSequenceTag();
        cmd[5] = 0x01;
        
        cmd.push_back(static_cast<uint8>(asic_mask));
        PrintCmd(cmd);
        
        m_tcp->SendData(cmd);

        vector<char> response(6,0x00);
        return CheckResponse(response);
    }

    RESPONSE AGIPDComm::SetTemperatureWarn(vector<int16>& value)
    {
        LOG_TRACE(__FUNCTION__);

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x81;
        cmd[1] = 0x09; // command code
        cmd[2] = 0x00; // to ZYNC

        cmd[3] = GetSequenceTag();
        cmd[5] = 0x03;
        
        cmd.push_back(static_cast<uint8>(value[0]));
        cmd.push_back(static_cast<uint8>(value[1]));
        cmd.push_back(static_cast<uint8>(value[2]));
        
        PrintCmd(cmd);
        
        m_tcp->SendData(cmd);

        vector<char> response(6,0x00);
        return CheckResponse(response);
    }

    RESPONSE AGIPDComm::SetTemperatureShutDown(vector<int16>& value)
    {
        LOG_TRACE(__FUNCTION__);

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x81;
        cmd[1] = 0x0a; // command code
        cmd[2] = 0x00; // to ZYNC

        cmd[3] = GetSequenceTag();
        cmd[5] = 0x03;
        
        cmd.push_back(static_cast<uint8>(value[0]));
        cmd.push_back(static_cast<uint8>(value[1]));
        cmd.push_back(static_cast<uint8>(value[2]));
        
        PrintCmd(cmd);
        
        m_tcp->SendData(cmd);

        vector<char> response(6,0x00);
        return CheckResponse(response);
    }

    RESPONSE AGIPDComm::SetInterlockMask(int8 mask)
    {
        LOG_TRACE(__FUNCTION__);

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x81;
        cmd[1] = 0x0b; // command code
        cmd[2] = 0x00; // to ZYNC

        cmd[3] = GetSequenceTag();
        cmd[5] = 0x01;
        
        cmd.push_back(static_cast<uint8>(mask));
        
        PrintCmd(cmd);
        
        m_tcp->SendData(cmd);

        vector<char> response(6,0x00);
        return CheckResponse(response);
    }

    RESPONSE AGIPDComm::GetTemperatures(vector<int16>& temperatures)
    {
        LOG_TRACE(__FUNCTION__);

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x80;
        cmd[1] = 0x0c; // command code
        cmd[2] = 0x00; // to ZYNC

        cmd[3] = GetSequenceTag();
        cmd[5] = 0x00;
        
        PrintCmd(cmd);
        
        m_tcp->SendData(cmd);

        vector<char> response(9,0x00);
        RESPONSE rc = CheckResponse(response);

        temperatures.resize(3,0);
        
        if(rc == SC)
        {
            temperatures[0] = response[6];
            temperatures[1] = response[7];
            temperatures[2] = response[8];
        }
        
        return rc;
            
    }

    RESPONSE AGIPDComm::GetPowerError(int16& powererr)
    {
        LOG_TRACE(__FUNCTION__);

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x80;
        cmd[1] = 0x0d; // command code
        cmd[2] = 0x00; // to ZYNC

        cmd[3] = GetSequenceTag();
        cmd[5] = 0x00;
        
        PrintCmd(cmd);
        
        m_tcp->SendData(cmd);
        

        vector<char> response(8,0x00);
        RESPONSE rc = CheckResponse(response);
        
        if(rc == SC)
            powererr = (((int16)response[6] << 8) | ((int16)response[7]));
        else
            powererr = -1;
        
        return rc;
    }

    RESPONSE AGIPDComm::SendExperimentFile(string& file_path, string& log)
    {

        LOG_TRACE(__FUNCTION__);

        if(!(m_fileop->FileExists(file_path)))
        {
            LOG_STREAM(__FUNCTION__,ERROR,"file doesn't exist : " + file_path);
            return UC;
        }
        
        string s = ReadFile(file_path);
        szt len = s.length();
        
        if(len == 0)
        {
            LOG_STREAM(__FUNCTION__,ERROR,"no data is sent");
            return UC;
        }

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x81;
        cmd[1] = 0x80; // command code
        cmd[2] = 0x00; // to ZYNC
        
        RESPONSE rc = SendBlob(cmd,s);
        if(rc != SC)
            return rc;
        
        vector<char> response(6,0x00);
        rc = CheckResponseLog(response);
        
        string content(response.begin(),response.end());
        log = content;

        //if(rc == SC)
        WriteLogToFile("compilation",content);
        
        return rc;
        
    }
    
    RESPONSE AGIPDComm::SetDataThreshold(string threshold_file)
    {
        LOG_TRACE(__FUNCTION__);

        if(!(m_fileop->FileExists(threshold_file)))
        {
            LOG_STREAM(__FUNCTION__,ERROR,"file doesn't exist : " + threshold_file);
            return UC;
        }

        string s = ReadBinaryFile(threshold_file);
        szt len = s.length();
        
        if(len == 0)
        {
            LOG_STREAM(__FUNCTION__,ERROR,"no data is sent");
            return UC;
        }

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x81;
        cmd[1] = 0x83; // command code
        cmd[2] = 0x00; // to ZYNC

        RESPONSE rc = SendBlob(cmd,s);

        if(rc == SC)
        {
            vector<char> response(6,0x00);
            return CheckResponse(response);
        }
        
        return rc;
    }

    RESPONSE AGIPDComm::StartAcq(string& log)
    {
        LOG_TRACE(__FUNCTION__);

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x80;
        cmd[1] = 0x81; // command code
        cmd[2] = 0x00; // to ZYNC

        cmd[3] = GetSequenceTag();
        cmd[5] = 0x00;
        
        PrintCmd(cmd);
        
        m_tcp->SendData(cmd);

        // vector<char> response(6,0x00);
        // return CheckResponse(response);

        vector<char> response(6,0x00);
        RESPONSE rc = CheckResponseLog(response);
        
        string content(response.begin(),response.end());
        log = content;

        // if(rc == SC)
        WriteLogToFile("run",content);
        
        return rc;
    }

    RESPONSE AGIPDComm::StopAcq()
    {
        LOG_TRACE(__FUNCTION__);

        vector<uint8> cmd = m_cmd;
        cmd[0] = 0x80;
        cmd[1] = 0x82; // command code
        cmd[2] = 0x00; // to ZYNC

        cmd[3] = GetSequenceTag();
        cmd[5] = 0x00;
        
        PrintCmd(cmd);
        
        m_tcp->SendData(cmd);

        vector<char> response(6,0x00);
        return CheckResponse(response);
    }

    string AGIPDComm::ReadFile(string& file_path)
    {
        LOG_TRACE(__FUNCTION__);
        
        m_fileop->SetFilePath(file_path);
        
        m_fileop->OpenFile(false);
        vector<string> data = m_fileop->ReadDataFromFile();
        m_fileop->CloseFile();

        string content;
        //content = accumulate(data.begin(),data.end(),content);
        
        char ending = 0x0a;
        for(auto &val : data)
        {
            if(val.find('\r') != std::string::npos)
            {
                LOG_STREAM(__FUNCTION__,ERROR,"Non linux line ending file " + file_path);
                content.clear();
                return content;
            }
            content = content + val + ending;
        }
        
        LOG_INFOS("length of content is:" + to_string(content.length()));
        
        return content;
    }

    string AGIPDComm::ReadBinaryFile(string& file_path)
    {
        LOG_TRACE(__FUNCTION__);
        
        m_fileop->SetFilePath(file_path);
        
        m_fileop->OpenFile(false);
        vector<int8> data = m_fileop->ReadDataFromCharBinaryFile();
        m_fileop->CloseFile();

        string content(data.begin(),data.end());
        //content = accumulate(data.begin(),data.end(),content);
        
        return content;
    }

    RESPONSE AGIPDComm::SendBlob(vector<uint8>& cmd,string& data)
    {
        LOG_TRACE(__FUNCTION__);

        int32 length = data.length();
        int32 remain = length;
        int32 sendsize = 0;
        char* ch = (char*)(data.c_str());
        
        while(remain > 0)
        {
            vector<uint8> cmdbase = cmd;
            cmdbase[3] = GetSequenceTag();

            if(remain >= MAX_PACKET_LENGTH)
            {
                cmdbase[4] = 0xff;
                cmdbase[5] = 0xff;

                cmdbase.insert(cmdbase.end(),ch,ch+MAX_PACKET_LENGTH);
                sendsize = MAX_PACKET_LENGTH;
            }
            else
            {
                cmdbase[4] = static_cast<uint8>((remain >> 8) & 0xff);
                cmdbase[5]= static_cast<uint8>(remain & 0xff);

                cmdbase.insert(cmdbase.end(),ch,ch+remain);
                sendsize = remain;
            }

            remain -= sendsize;
            ch += sendsize;

            PrintCmd(cmdbase);
            
            m_tcp->SendData(cmdbase);
            cmdbase.clear();

            vector<char> response(6,0x00);
            RESPONSE rc = CheckResponse(response);
            if(rc != SC)
                return rc;
        }

        //send last command
        cmd[3] = GetSequenceTag();
        PrintCmd(cmd);
        m_tcp->SendData(cmd);

        return SC;
    }
    
    uint8 AGIPDComm::GetSequenceTag()
    {
        LOG_TRACE(__FUNCTION__);
        
        return static_cast<uint8>(((m_tag++ == 255)?0:m_tag) & 0xff);
    }

    void AGIPDComm::PrintCmd(vector<uint8>& cmd)
    {
        LOG_STREAM1(__FUNCTION__,INFO,cmd);
    }
    
    RESPONSE AGIPDComm::CheckResponse(vector<char>& response)
    {
        LOG_TRACE(__FUNCTION__);
        
        szt rec_data_len = 0;
        char* rec_data = response.data();
        szt current_read = 0;
        szt requested_size = response.size();
        
        int32 ret = m_tcp->ReceiveData(rec_data,requested_size,requested_size,current_read);
        
        vector<uint8> data(response.size());
        std::copy(response.begin(),response.end(),data.begin());
        
        PrintCmd(data);
        
        if(ret == -1)
        {
            LOG_STREAM(__FUNCTION__,ERROR,"hardware response time out");
            return UC;
        }
        else
        {
            LOG_INFOS("Received data : " + to_string(rec_data_len));
            return static_cast<RESPONSE>(response[1]);
        }
        
        // LOG_INFOS("Received data : " + to_string(rec_data_len));
        
        // return SC;
    }

    RESPONSE AGIPDComm::CheckResponseLog(vector<char>& response)
    {
        LOG_TRACE(__FUNCTION__);
        
        //szt rec_data_len = 0;
        RESPONSE rs;
        char* rec_data = response.data();
        szt current_read = 0;
        szt requested_size = response.size();
        
        int32 ret = m_tcp->ReceiveData(rec_data,requested_size,requested_size,current_read);

        if(ret == -1)
        {
            LOG_STREAM(__FUNCTION__,ERROR,"hardware response time out");
            return UC;
        }
        else
        {
            LOG_INFOS("get response:" + static_cast<int>(response[1]));
            rs = static_cast<RESPONSE>(response[1]);
        }
        

        vector<uint8> data(response.size());
        std::copy(response.begin(),response.end(),data.begin());
        
        PrintCmd(data);
        
        uint16 data_length
            = ((static_cast<uint16>(response[4]) << 8) | (static_cast<uint16>(response[5])& 0xff));

        if(data_length < 0)
        {
            LOG_STREAM(__FUNCTION__,ERROR,"data length error : " + to_string(data_length));
            return UC;
        }

        LOG_INFOS("Expected data : " + to_string(data_length));


        response.resize(data_length);
        rec_data = response.data();
        requested_size = response.size();
        current_read = 0;
        
        ret = m_tcp->ReceiveData(rec_data,requested_size,requested_size,current_read);

        if(ret == -1)
        {
            LOG_STREAM(__FUNCTION__,ERROR,"hardware response time out. Received:"+to_string(current_read));
            return UC;
        }

        data.clear();
        
        data = vector<uint8>(response.size());
        std::copy(response.begin(),response.end(),data.begin());
        
        PrintCmd(data);
        
        LOG_INFOS("Received data : " + to_string(response.size()));
        
        return rs;
    }

    void AGIPDComm::WriteLogToFile(string file_prefix, string& log)
    {
        LOG_TRACE(__FUNCTION__);

        string filename = m_logpath + "/" + file_prefix+"_" + GetTimeByString() + ".log";
        LOG_INFOS("filename : " + filename);

        std::ofstream out(filename,std::ofstream::out);
        out << log;
        if(out.fail())
            LOG_STREAM(__FUNCTION__,ERROR,"error occurred, cannot write log to file" + filename);
        out.close();
        LOG_INFOS("write log to file : " + filename);

        
    }

    string AGIPDComm::GetTimeByString()
    {
        LOG_TRACE(__FUNCTION__);
        std::time_t t = std::time(nullptr);
        struct tm * timeinfo;
        timeinfo = localtime ( &t );

        static char result[26];
        //display at least 4 digits for year and 2 digits for others
        // .precision should be used, otherwise 01 will be like 1
        sprintf(result, "%.4d%.2d%.2d_%.2d%.2d%.2d",     
                1900 + timeinfo->tm_year,
                timeinfo->tm_mon+1,
                timeinfo->tm_mday,
                timeinfo->tm_hour,
                timeinfo->tm_min,
                timeinfo->tm_sec);
        
        return string(result);

    }
}
