/*
 * (c) Copyright 2018-2020 DESY
 *
 * This file is part of FS-DS detector software.
 *
 * This software is free: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 *************************************************************************
 *     Author: Yuelong Yu <yuelong.yu@desy.de>
 */

#pragma once

#include "AGIPDCtrlGlobals.h"

namespace AGIPDCtrlNS
{
    // class AGIPDComm ////////////////////////////////////////////////////////
    /**
     *  \brief AGIPD communication class
     *  The class encapsulates the communication commands
     *  between PC and detector
     */
    class AGIPDComm
    {
    public:
        /**
         *  \brief constructor
         */
        AGIPDComm();
        
        /**
         *  \brief destructor
         */
        ~AGIPDComm();

        /**
         *  \brief init the communication part
         *  \param tcp connection for communication
         *  \param log_path path for saving log files
         *  \return status. true : OK
         *                  flase : init is failed
         */
        bool Init(sptr_network& tcp,string log_path);

        /**
         *  \brief get firmware version
         *  \return firmware version
         */
        string GetFirmwareVersion();

        /**
         *  \brief get application log
         *  \param log log information of the hardware
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE GetApplicationLog(string& log);

        /**
         *  \brief get application log
         *  \param log log information of the hardware
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE GetCompilationLog(string& log);

         /**
         *  \brief set 10GE link MAC
         *  \param mac mac address
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE Set10GElinkMAC(string mac);

        /**
         *  \brief set 10GE link IP
         *  \param ip ip address
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE Set10GElinkIP(string ip);
        

        /**
         *  \brief set high voltage
         *  \param high_voltage 0-65535
         *  \return response of command \see AGIPDCtrlNS::RESPONSE        
         */
        RESPONSE SetHighVoltage(uint16 high_voltage);
        
        /**
         *  \brief get high voltage
         *  \param  high_voltage high voltage
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE GetHighVoltage(uint16& high_voltage);

        /**
         *  \brief set high voltage slew rate
         *  \param hv_slewrate high voltage slew rate,0-65535
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE SetHighVoltageSlewRate(uint16 hv_slewrate);

        /**
         *  \brief get high voltage slew rate
         *  \param hv_slewrate high voltage slew rate
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE GetHighVoltageSlewRate(uint16& hv_slewrate);
        
        /**
         *  \brief power up ASIC
         *  \param asic_mask 1 byte 0-255 each bit in the byte represents two ASICS from 16
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE ASICPowerUP(int8 asic_mask);

        /**
         *  \brief power down asic
         *  \param asic_mask 1 byte 0-255 each bit in the byte represents two ASICS from 16
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE ASICPowerDown(int8 asic_mask);

        /**
         *  \brief set temperature threshold for system warnning
         *  \param value threshold of temperature
         *               it contains three values, which set the threshold
         *               for three temperature sensors respectively.
         *               \see  AGIPDCtrlNS::TEMPERATURE_SENSOR
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE SetTemperatureWarn(vector<int16>& value);

        /**
         *  \brief set temperature threshold for system shut down
         *  \param value threshold of temperature
         *               it contains three values, which set the threshold
         *               for three temperature sensors respectively.
         *               \see  AGIPDCtrlNS::TEMPERATURE_SENSOR
         *  \return response of command \see AGIPDCtrlNS::RESPONSE             
         */
        RESPONSE SetTemperatureShutDown(vector<int16>& value);

        /**
         *  \brief set interlock mask
         *  \param mask mask is 3bits
         *              000 : no mask
         *              001 : mask LTCC
         *              010 : mask ZYNC
         *              100 : mask FF
         *  \return response of command \see AGIPDCtrlNS::RESPONSE  
         */
        RESPONSE SetInterlockMask(int8 mask);

        /**
         *  \brief get temperature
         *  \param temperatures list temperatures corresponding to different temprature sensors
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE GetTemperatures(vector<int16>& temperatures);

        /**
         *  \brief get power error
         *  \param powererr each bit reports failure power supply
         *          0x00 : POK
         *  \return response of command \see AGIPDCtrlNS::RESPONSE         
         */
        RESPONSE GetPowerError(int16& powererr);

        /**
         *  \brief send experiment file
         *  \param file_path experiment file path
         *  \param log compilation log
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE SendExperimentFile(string& file_path,string& log);

         /**
         *  \brief set data threshold
         *  \param threshold file
         *  \return response of command \see AGIPDCtrlNS::RESPONSE 
         */
        RESPONSE SetDataThreshold(string threshold_file);
        
        /**
         *  \brief start acquisition
         *  \param log run log
         *  \return response of command \see AGIPDCtrlNS::RESPONSE 
         */
        RESPONSE StartAcq(string& log);

        /**
         *  \brief stop acquisition
         *  \return response of command \see AGIPDCtrlNS::RESPONSE 
         */
        RESPONSE StopAcq();
        
    private:
        /**
         *  \brief read file content
         *  \param file_path file path
         *  \return content of the file
         */
        string ReadFile(string& file_path);

        /**
         *  \brief read binary file content
         *  \param file_path file path
         *  \return content of the file
         */
        string ReadBinaryFile(string& file_path);
        
        /**
         *  \brief get sequence tag for the commands, the range is 0-255
         *  \return sequence tag
         */
        uint8 GetSequenceTag();

        /**
         *  \brief send blob
         *  \param cmd base command
         *  \param data whole data
         *  \return response of command \see AGIPDCtrlNS::RESPONSE 
         */
        RESPONSE  SendBlob(vector<uint8>& cmd,string& data);

        /**
         *  \brief print command
         *  \param cmd
         */
        void PrintCmd(vector<uint8>& cmd);

        /**
         *  \brief check response of the hardware
         *  \param response the data returned from hardware
         *  \return response of the command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE CheckResponse(vector<char>& response);

         /**
         *  \brief check response of the hardware
         *  \param response the data returned from hardware
         *  \return response of the command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE CheckResponseLog(vector<char>& response);

        /**
         *  \brief write log to file
         *  \param file_prefix file prefix
         *  \param log log
         */
        void WriteLogToFile(string file_prefix, string& log);

        /**
         *  \brief get time by string
         *  \return time
         */
        string GetTimeByString();
        
        // member variables
        sptr_network m_tcp;
        sptr_fileop m_fileop;
        
        vector<uint8> m_cmd;
        vector<uint8> m_response;

        // sequence tag for command
        int32 m_tag;

        string m_logpath;
    };
    
    
}
