/*
 * (c) Copyright 2018-2020 DESY
 *
 * This file is part of FS-DS detector software.
 *
 * This software is free: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 *************************************************************************
 *     Author: Yuelong Yu <yuelong.yu@desy.de>
 */

#include "AGIPDCtrlInterface.h"
#include "AGIPDComm.h"

namespace AGIPDCtrlNS
{
   
    AGIPDCtrlInterface::AGIPDCtrlInterface(string ip, int16 port,Enum_log_level level)
        :m_ip(ip),
         m_port(port),
         m_state(ON)
    {
        FSDetCoreNS::InitLogLevel(level);
        
        LOG_TRACE(__FUNCTION__);
    }
    
    AGIPDCtrlInterface::~AGIPDCtrlInterface()
    {
        LOG_TRACE(__FUNCTION__);
    }
    
    bool AGIPDCtrlInterface::InitAGIPDCtrl(string log_path)
    {
        LOG_TRACE(__FUNCTION__);
        
        m_network
            = sptr_network(new NetworkTCPInterface(new NetworkTCPImplementation(m_ip,m_port)));

        if(m_network->Connect() != 0)
        {
            LOG_STREAM(__FUNCTION__,ERROR,"TCP init failed");
            
            return false;
        }

        m_communication = uptr_agipdcomm(new AGIPDComm());
        if(!(m_communication->Init(m_network,log_path)))
        {
            LOG_STREAM(__FUNCTION__,ERROR,"detector control init failed");
            return false;
        }

        return true;
    }
    
    string AGIPDCtrlInterface::GetDetCoreVersion()
    {
        LOG_TRACE(__FUNCTION__);

        return FSDetCoreNS::FSDETCORE_VERSION;
    }

    string AGIPDCtrlInterface::GetLibAGIPDCtrlVersion()
    {
        LOG_TRACE(__FUNCTION__);

        return AGIPDCTRL_VERSION;
    }

    string AGIPDCtrlInterface::GetFirmwareVersion()
    {
        LOG_TRACE(__FUNCTION__);

        return m_communication->GetFirmwareVersion();
    }
    
    Enum_detector_state AGIPDCtrlInterface::GetStatus()
    {
        LOG_TRACE(__FUNCTION__);

        return m_state;
    }

    RESPONSE AGIPDCtrlInterface::SendExperimentFile(string& file_path,string& log)
    {
        LOG_TRACE(__FUNCTION__);

        return m_communication->SendExperimentFile(file_path,log);
    }


    RESPONSE AGIPDCtrlInterface::GetApplicationLog(string& log)
    {
        LOG_TRACE(__FUNCTION__);
        
        return m_communication->GetApplicationLog(log);
    }
    
    RESPONSE AGIPDCtrlInterface::GetCompilationLog(string& log)
    {
        LOG_TRACE(__FUNCTION__);
        
        return m_communication->GetCompilationLog(log);
    }

    RESPONSE AGIPDCtrlInterface::Set10GElinkMAC(string mac)
    {
        LOG_TRACE(__FUNCTION__);

        return m_communication->Set10GElinkMAC(mac);
    }
    
    RESPONSE AGIPDCtrlInterface::Set10GElinkIP(string ip)
    {
        LOG_TRACE(__FUNCTION__);
        
        return m_communication->Set10GElinkIP(ip);
    }

    RESPONSE AGIPDCtrlInterface::SetHighVoltage(uint16 high_voltage)
    {
        LOG_TRACE(__FUNCTION__);

        return m_communication->SetHighVoltage(high_voltage);
    }
    
    RESPONSE AGIPDCtrlInterface::GetHighVoltage(uint16& high_voltage)
    {
        LOG_TRACE(__FUNCTION__);
        
        return m_communication->GetHighVoltage(high_voltage);
    }

    RESPONSE AGIPDCtrlInterface::SetHighVoltageSlewRate(uint16 hv_slewrate)
    {
        LOG_TRACE(__FUNCTION__);
        
        return m_communication->SetHighVoltageSlewRate(hv_slewrate);
    }

    RESPONSE AGIPDCtrlInterface::GetHighVoltageSlewRate(uint16& hv_slewrate)
    {
        LOG_TRACE(__FUNCTION__);

        return m_communication->GetHighVoltageSlewRate(hv_slewrate);
    }
    
    RESPONSE AGIPDCtrlInterface::ASICPowerUP(int8 asic_mask)
    {
        LOG_TRACE(__FUNCTION__);

        return m_communication->ASICPowerUP(asic_mask);
    }
    
    RESPONSE AGIPDCtrlInterface::ASICPowerDown(int8 asic_mask)
    {
        LOG_TRACE(__FUNCTION__);

        return m_communication->ASICPowerDown(asic_mask);
    }
      
    RESPONSE AGIPDCtrlInterface::SetTemperatureWarn(vector<int16>& value)
    {
        LOG_TRACE(__FUNCTION__);

        return m_communication->SetTemperatureWarn(value);
    }
    
    RESPONSE AGIPDCtrlInterface::SetTemperatureShutDown(vector<int16>& value)
    {
        LOG_TRACE(__FUNCTION__);

        return m_communication->SetTemperatureShutDown(value);
    }

    RESPONSE AGIPDCtrlInterface::SetInterlockMask(int8 mask)
    {
        LOG_TRACE(__FUNCTION__);

        return m_communication->SetInterlockMask(mask);
    }
    
    RESPONSE AGIPDCtrlInterface::GetTemperatures(vector<int16>& temperatures)
    {
        LOG_TRACE(__FUNCTION__);

        return m_communication->GetTemperatures(temperatures);
    }
    
    RESPONSE AGIPDCtrlInterface::GetPowerError(int16& powererr)
    {
        LOG_TRACE(__FUNCTION__);

        return m_communication->GetPowerError(powererr);
    }

    RESPONSE AGIPDCtrlInterface::SetDataThreshold(string threshold_file)
    {
        LOG_TRACE(__FUNCTION__);

        return m_communication->SetDataThreshold(threshold_file);
    }
  
    RESPONSE AGIPDCtrlInterface::StartAcq(string& log)
    {
        LOG_TRACE(__FUNCTION__);

        return m_communication->StartAcq(log);
    }
    
    RESPONSE AGIPDCtrlInterface::StopAcq()
    {
        LOG_TRACE(__FUNCTION__);

        return m_communication->StopAcq();
    }

    string AGIPDCtrlInterface::GetResponseType(RESPONSE rc)
    {
        LOG_TRACE(__FUNCTION__);

        switch(rc)
        {
            case 0:
                return string("successful compeletion");
            case 1:
                return string("unsupported command");
            case 2:
                return string("unsupported device");
            case 3:
                return string("compilation error");
            default:
                return string("unknown response type");
        }
    }
}
