/*
 * (c) Copyright 2018-2020 DESY
 *
 * This file is part of FS-DS detector software.
 *
 * This software is free: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 *************************************************************************
 *     Author: Yuelong Yu <yuelong.yu@desy.de>
 */

#pragma once

#include "AGIPDCtrlGlobals.h"

namespace AGIPDCtrlNS
{
    class AGIPDComm;
    
    typedef unique_ptr<AGIPDComm> uptr_agipdcomm;

    // class AGIPDCtrlInterface ///////////////////////////////////////////////
    /**
     *  \brief AGIPDCtrlInterface
     *  This class defines the interface of detector
     */
    class AGIPDCtrlInterface
    {
    public:
        /**
         *  \brief constructor
         *  \param ip ip address of detector
         *  \param port communication port
         *  \param level log level. default is ERROR see FSDetCoreNS::Enum_log_level
         */
        AGIPDCtrlInterface(string ip, int16 port,Enum_log_level level = ERROR);

        /**
         *  \brief destructor
         */
        ~AGIPDCtrlInterface();

        /**
         *  \brief init agipd control
         *  \param log_path path for saving log files
         *  \return true : OK
         *          false : error occurred
         */
        bool InitAGIPDCtrl(string log_path=string("./"));

        /**
         *  \brief get core library version
         *  \return the core library version
         */
        string GetDetCoreVersion();

        /**
         *  \brief get agipd control library version
         *  \return the agipd control library version
         */
        string GetLibAGIPDCtrlVersion();

        /**
         *  \brief get firmware version
         *  \return firmware version
         */
        string GetFirmwareVersion();

        /**
         *  \brief get status
         *           
         *  For details of type of the status
         *  \see FSDetCoreNS::Enum_detector_state
         *  
         *  \return return status of the system
         */
        Enum_detector_state GetStatus();
        
        /**
         *  \brief send experiment file
         *  \param file_path path of experiment file
         *  \param log compilation log
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE SendExperimentFile(string& file_path,string& log);

        /**
         *  \brief get application log
         *  \param log log information of the hardware
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE GetApplicationLog(string& log);

        /**
         *  \brief get application log
         *  \param log log information of the hardware
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE GetCompilationLog(string& log);

        /**
         *  \brief set 10GE link MAC
         *  \param mac mac address
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE Set10GElinkMAC(string mac);

        /**
         *  \brief set 10GE link IP
         *  \param ip ip address
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE Set10GElinkIP(string ip);

        /**
         *  \brief set high voltage
         *  \param high_voltage 0-65535
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE SetHighVoltage(uint16 high_voltage);

        /**
         *  \brief get high voltage
         *  \param  high_voltage high voltage
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE GetHighVoltage(uint16& high_voltage);

        /**
         *  \brief set high voltage slew rate
         *  \param hv_slewrate high voltage slew rate,0-65535
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE SetHighVoltageSlewRate(uint16 hv_slewrate);

        /**
         *  \brief get high voltage slew rate
         *  \param hv_slewrate high voltage slew rate
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE GetHighVoltageSlewRate(uint16& hv_slewrate);
        
        /**
         *  \brief power up ASIC
         *  \param asic_mask 1 byte 0-255 each bit in the byte represents two ASICS from 16
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE ASICPowerUP(int8 asic_mask);

        /**
         *  \brief power down asic
         *  \param asic_mask 1 byte 0-255 each bit in the byte represents two ASICS from 16
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE ASICPowerDown(int8 asic_mask);

        /**
         *  \brief set temperature threshold for system warnning
         *  \param value threshold of temperature
         *               it contains three values, which set the threshold
         *               for three temperature sensors respectively.
         *               \see  AGIPDCtrlNS::TEMPERATURE_SENSOR
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE SetTemperatureWarn(vector<int16>& value);

        /**
         *  \brief set temperature threshold for system shut down
         *  \param value threshold of temperature
         *               it contains three values, which set the threshold
         *               for three temperature sensors respectively.
         *               \see  AGIPDCtrlNS::TEMPERATURE_SENSOR
         *  \return response of command \see AGIPDCtrlNS::RESPONSE             
         */
        RESPONSE SetTemperatureShutDown(vector<int16>& value);

        /**
         *  \brief set interlock mask
         *  \param mask mask is 3bits
         *              000 : no mask
         *              001 : mask LTCC
         *              010 : mask ZYNC
         *              100 : mask FF
         *  \return response of command \see AGIPDCtrlNS::RESPONSE  
         */
        RESPONSE SetInterlockMask(int8 mask);

        /**
         *  \brief get temperature
         *  \param temperatures list temperatures corresponding to different temprature sensors
         *  \return response of command \see AGIPDCtrlNS::RESPONSE
         */
        RESPONSE GetTemperatures(vector<int16>& temperatures);

        /**
         *  \brief get power error
         *  \param powererr each bit reports failure power supply
         *          0x00 : POK
         *  \return response of command \see AGIPDCtrlNS::RESPONSE         
         */
        RESPONSE GetPowerError(int16& powererr);

         /**
         *  \brief set data threshold
         *  \param threshold file
         *  \return response of command \see AGIPDCtrlNS::RESPONSE 
         */
        RESPONSE SetDataThreshold(string threshold_file);
        
        /**
         *  \brief start acquisition
         *  \return response of command \see AGIPDCtrlNS::RESPONSE 
         */
        RESPONSE StartAcq(string& log);

        /**
         *  \brief stop acquisition
         *  \return response of command \see AGIPDCtrlNS::RESPONSE 
         */
        RESPONSE StopAcq();

        /**
         *  \brief get response string via response type
         *  \param rc response type
         *  \string error information
         */
        string GetResponseType(RESPONSE rc);
        
    private:
        string m_ip;
        int16 m_port;
        Enum_detector_state m_state;

        sptr_network m_network;
        uptr_agipdcomm m_communication;
    };
    
    
}
