/*
 * (c) Copyright 2018-2020 DESY
 *
 * This file is part of FS-DS detector software.
 *
 * This software is free: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 *************************************************************************
 *     Author: Yuelong Yu <yuelong.yu@desy.de>
 */

#include "CtrlTest.h"

namespace TestNS
{
    void CtrlTest::SetUp()
    {
        InitLogLevel(TRACE);

        m_testdir = "./testdata/";
       
        m_tcp =sptr_mocknet(new MockNetwork());
        m_agipd = uptr_agipd(new AGIPDComm());
        m_file =uptr_file(new FileReader());

        auto ptr = static_pointer_cast<NetworkInterface>(m_tcp);
        EXPECT_TRUE(m_agipd->Init(ptr,m_testdir));
    }
    
    void CtrlTest::TearDown()
    {
        m_agipd.reset();
        m_file.reset();
        m_output.clear();
    }

    TEST_F(CtrlTest,GetFirmwareVersion)
    {
        EXPECT_EQ("unknown",m_agipd->GetFirmwareVersion());
    }

    TEST_F(CtrlTest,GetApplicationLog)
    {
        string val;
        vector<uint8> ref_data={0x80,0x03,0x00,0x01,0x00,0x00};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x0,0x2};
        vector<uint8> data_ret = {0x48,0x49};//HI
        string ref_return(data_ret.begin(),data_ret.end());
        
        szt length = response.size();
        szt data_length = data_ret.size();
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,_,_,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_CALL(*m_tcp,ReceiveData(_,data_length,data_length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(data_ret.begin(),data_ret.end()),
                            SetArgReferee<3>(data_length),
                            Return(0)));
        
        EXPECT_EQ(SC,m_agipd->GetApplicationLog(val));
        EXPECT_EQ(ref_data,m_output);
        EXPECT_EQ(ref_return,val);
    }
    
    TEST_F(CtrlTest,GetCompilationLog)
    {
        string val;
        vector<uint8> ref_data={0x80,0x02,0x00,0x01,0x00,0x00};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x0,0x2};
        vector<uint8> data_ret = {0x48,0x49};//HI
        string ref_return(data_ret.begin(),data_ret.end());
        
        szt length = response.size();
        szt data_length = data_ret.size();
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,_,_,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_CALL(*m_tcp,ReceiveData(_,data_length,data_length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(data_ret.begin(),data_ret.end()),
                            SetArgReferee<3>(data_length),
                            Return(0)));
        
        EXPECT_EQ(SC,m_agipd->GetCompilationLog(val));
        EXPECT_EQ(ref_data,m_output);
        EXPECT_EQ(ref_return,val);
    }

    TEST_F(CtrlTest,Set10GElinkMAC)
    {
        string mac="aa:bb:cc:dd:ee:ff";
        vector<uint8> ref_data={0x81,0x00,0x00,0x01,0x00,0x06,0xaa,0xbb,0xcc,0xdd,0xee,0xff};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6};
        
        szt length = response.size();
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,length,length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_EQ(SC,m_agipd->Set10GElinkMAC(mac));
        EXPECT_EQ(ref_data,m_output);
    }

    TEST_F(CtrlTest,Set10GElinkIP)
    {
        string ip="127.0.0.1";
        vector<uint8> ref_data={0x81,0x01,0x00,0x01,0x00,0x04,0x7f,0x00,0x00,0x01};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6};
        
        szt length = response.size();
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,length,length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_EQ(SC,m_agipd->Set10GElinkIP(ip));
        EXPECT_EQ(ref_data,m_output);
    }

    TEST_F(CtrlTest,SetHighVoltage)
    {
        uint16 high_vol = 1234;
        vector<uint8> ref_data={0x81,0x04,0x00,0x01,0x00,0x02,0x04,0xd2};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6};
        
        szt length = response.size();
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,length,length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_EQ(SC,m_agipd->SetHighVoltage(high_vol));
        EXPECT_EQ(ref_data,m_output);
    }

    TEST_F(CtrlTest,GetHighVoltage)
    {
        uint16 high_vol;
        vector<uint8> ref_data={0x80,0x05,0x00,0x01,0x00,0x00};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6,0xa,0xa};
        
        szt length = response.size();
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,length,length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_EQ(SC,m_agipd->GetHighVoltage(high_vol));
        EXPECT_EQ(ref_data,m_output);
        EXPECT_EQ(high_vol,(((int16)response[6] << 8) | ((int16)response[7])));
    }

    TEST_F(CtrlTest,SetHighVoltageSlewRate)
    {
        uint16 high_vol_slew_rate = 1234;
        vector<uint8> ref_data={0x81,0x06,0x00,0x01,0x00,0x02,0x04,0xd2};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6};
        
        szt length = response.size();
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,length,length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_EQ(SC,m_agipd->SetHighVoltageSlewRate(high_vol_slew_rate));
        EXPECT_EQ(ref_data,m_output);
    }

    TEST_F(CtrlTest,GetHighVoltageSlewRate)
    {
        uint16 high_vol_slew_rate;
        vector<uint8> ref_data={0x80,0x05,0x00,0x01,0x00,0x00};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6,0xa,0xa};
        
        szt length = response.size();
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,length,length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_EQ(SC,m_agipd->GetHighVoltageSlewRate(high_vol_slew_rate));
        EXPECT_EQ(ref_data,m_output);
        EXPECT_EQ(high_vol_slew_rate,(((int16)response[6] << 8) | ((int16)response[7])));
    }

    TEST_F(CtrlTest,ASICPowerUP)
    {
        int8 mask = 0x07;
        vector<uint8> ref_data={0x80,0x07,0x00,0x01,0x00,0x01,0x07};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6};
        
        szt length = response.size();
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,length,length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_EQ(SC,m_agipd->ASICPowerUP(mask));
        EXPECT_EQ(ref_data,m_output);
    }

    TEST_F(CtrlTest,ASICPowerDown)
    {
        int8 mask = 0x07;
        vector<uint8> ref_data={0x80,0x08,0x00,0x01,0x00,0x01,0x07};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6};
        
        szt length = response.size();
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,length,length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_EQ(SC,m_agipd->ASICPowerDown(mask));
        EXPECT_EQ(ref_data,m_output);
    }

    TEST_F(CtrlTest,SetTemperatureWarn)
    {
        vector<int16> temperatures = {10,11,12};
        vector<uint8> ref_data={0x81,0x09,0x00,0x01,0x00,0x03,0x0a,0x0b,0x0c};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6};
        
        szt length = response.size();
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,length,length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_EQ(SC,m_agipd->SetTemperatureWarn(temperatures));
        EXPECT_EQ(ref_data,m_output);
    }

    TEST_F(CtrlTest,SetTemperatureShutDown)
    {
        vector<int16> temperatures = {50,51,52};
        vector<uint8> ref_data={0x81,0x0a,0x00,0x01,0x00,0x03,0x32,0x33,0x34};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6};
        
        szt length = response.size();
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,length,length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_EQ(SC,m_agipd->SetTemperatureShutDown(temperatures));
        EXPECT_EQ(ref_data,m_output);
    }

    TEST_F(CtrlTest,SetInterlockMask)
    {
        int8 interlock_mask = 0x07;
        vector<uint8> ref_data={0x81,0x0b,0x00,0x01,0x00,0x01,0x07};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6};
        
        szt length = response.size();
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,length,length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_EQ(SC,m_agipd->SetInterlockMask(interlock_mask));
        EXPECT_EQ(ref_data,m_output);
    }

    TEST_F(CtrlTest,GetTemperatures)
    {
        vector<int16> temperatures;
        vector<uint8> ref_data={0x80,0x0c,0x00,0x01,0x00,0x00};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6,0xa,0xb,0xc};
        
        szt length = response.size();
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,length,length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_EQ(SC,m_agipd->GetTemperatures(temperatures));
        EXPECT_EQ(ref_data,m_output);
    }
    
    TEST_F(CtrlTest,GetPowerError)
    {
        int16 powererr;
        vector<uint8> ref_data={0x80,0x0d,0x00,0x01,0x00,0x00};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6,0x7,0x8};
        
        szt length = response.size();
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,length,length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_EQ(SC,m_agipd->GetPowerError(powererr));
        EXPECT_EQ(ref_data,m_output);
        EXPECT_EQ(powererr,(((int16)response[6] << 8) | ((int16)response[7])));
    }

    TEST_F(CtrlTest,SendExperimentFile)
    {
        string expfile = m_testdir + "/experiment.c";
        string refexpfile = m_testdir + "/ref_exp.c";
        
        m_expfilelen = 2567; //bytes
        
        //2493 = 0x09 0xbd
        vector<uint8> ref_data={0x81,0x80,0x00,0x01,0x09,0xbd};
        vector<uint8> ref_dataend={0x81,0x80,0x00,0x02,0x00,0x00};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6};
        vector<uint8> responselog={0x0,0x0,0x3,0x4,0x0,0x2};
        vector<uint8> data_ret = {0x48,0x49};//HI
        string ref_return(data_ret.begin(),data_ret.end());
        szt lengthlog = responselog.size();
        szt data_length = data_ret.size();
        szt length = response.size();
        vector<uint8> output1;
        string log;

        m_file->SetFilePath(refexpfile);
        m_file->OpenFile(false);
        vector<string> data_from_file = m_file->ReadDataFromFile();
        m_file->CloseFile();

        vector<uint8> ref_dataoutput;
        for(auto val : data_from_file)
            ref_dataoutput.push_back(atoi(val.c_str()));
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(2)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)))
            .WillOnce(DoAll(SaveArg<0>(&output1),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,_,_,_))
            .Times(3)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                           SetArgReferee<3>(length),
                           Return(0)))
            .WillOnce(DoAll(SetArrayArgument<0>(responselog.begin(),responselog.end()),
                           SetArgReferee<3>(lengthlog),
                           Return(0)))
            .WillOnce(DoAll(SetArrayArgument<0>(data_ret.begin(),data_ret.end()),
                           SetArgReferee<3>(data_length),
                            Return(0)));
        
        
        EXPECT_EQ(SC,m_agipd->SendExperimentFile(expfile,log));
        
        EXPECT_EQ(m_expfilelen + ref_data.size(),m_output.size());
        EXPECT_EQ(ref_dataoutput,m_output);
        EXPECT_EQ(ref_dataend,output1);
        EXPECT_EQ(ref_return,log);

        // for(auto &val : m_output)
        //     cout<<static_cast<int16>(val)<<endl;

        // cout<<endl;
    }

    TEST_F(CtrlTest,SendExperimentFileFailWithWindowsEndingFile)
    {
        string expfile = m_testdir + "/windowsending.txt";
         vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6};
        szt length = response.size();
        string log;
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(0);
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,length,length,_))
            .Times(0);
        
        EXPECT_EQ(UC,m_agipd->SendExperimentFile(expfile,log));
    }
    
    TEST_F(CtrlTest,SetDataThreshold)
    {
        string thresfile = m_testdir + "/reduction.bin";
        string ref_thresfile = m_testdir + "/ref_reduction.txt";
        
        
        vector<uint8> ref_data={0x81,0x83,0x00,0x01,0x09,0xbd};
        vector<uint8> ref_dataend={0x81,0x81,0x00,0x02,0x00,0x00};

        //ref_data length + file length (131072)+ ref_dataend length
        szt threshold_file_len = 131096; 
        
        vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6};
        szt length = response.size();

        m_file->SetFilePath(ref_thresfile);
        m_file->OpenFile(false);
        vector<string> data_from_file = m_file->ReadDataFromFile();
        m_file->CloseFile();

        vector<uint8> ref_dataoutput;
        for(auto val : data_from_file)
            ref_dataoutput.push_back(atoi(val.c_str()));

        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(AtLeast(1))
            .WillRepeatedly(DoAll(Invoke(this,&CtrlTest::OutputData),Return(0)));

        EXPECT_CALL(*m_tcp,ReceiveData(_,length,length,_))
            .Times(AtLeast(1))
            .WillRepeatedly(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));
        
        EXPECT_EQ(SC,m_agipd->SetDataThreshold(thresfile));
        EXPECT_EQ(threshold_file_len,m_output.size());
        EXPECT_EQ(ref_dataoutput,m_output);

        // for(auto &val : m_output)
        //     cout<<static_cast<int16>(val)<<endl;

        // cout<<endl;
    }

    TEST_F(CtrlTest,StartAcq)
    {
        vector<uint8> ref_data={0x80,0x81,0x00,0x01,0x00,0x00};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x0,0x2};
        vector<uint8> data_ret = {0x48,0x49};//HI
        string ref_return(data_ret.begin(),data_ret.end());
        
        szt length = response.size();
        szt data_length = data_ret.size();
        
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));

        EXPECT_CALL(*m_tcp,ReceiveData(_,_,_,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_CALL(*m_tcp,ReceiveData(_,data_length,data_length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(data_ret.begin(),data_ret.end()),
                            SetArgReferee<3>(data_length),
                            Return(0)));
        string log;
        
        EXPECT_EQ(SC,m_agipd->StartAcq(log));
        EXPECT_EQ(ref_data,m_output);
        EXPECT_EQ(ref_return,log);
    }

    TEST_F(CtrlTest,StopAcq)
    {
        vector<uint8> ref_data={0x80,0x82,0x00,0x01,0x00,0x00};
        vector<uint8> response={0x0,0x0,0x3,0x4,0x5,0x6};
        
        szt length = response.size();
        
        EXPECT_CALL(*m_tcp,SendData(_))
            .Times(1)
            .WillOnce(DoAll(SaveArg<0>(&m_output),Return(0)));
        
        EXPECT_CALL(*m_tcp,ReceiveData(_,length,length,_))
            .Times(1)
            .WillOnce(DoAll(SetArrayArgument<0>(response.begin(),response.end()),
                            SetArgReferee<3>(length),
                            Return(0)));

        EXPECT_EQ(SC,m_agipd->StopAcq());
        EXPECT_EQ(ref_data,m_output);
    }

    void CtrlTest::OutputData(vector<uint8> data)
    {
        m_output.insert(m_output.end(),data.begin(),data.end());
    }
}